"""chem_objects.py
This file contains most of the classes provided by chanim. Includes ChemObject, Reaction, BondBreak etc.
"""

from typing import List, Union

import numpy as np

from .templates import ChemReactionTemplate, ChemTemplate
from .utils import Arrows

from manim.mobject.types.vectorized_mobject import VGroup, VMobject
from manim.mobject.svg.tex_mobject import *
from manim.animation.animation import Animation
from manim.animation.creation import Write
from manim.animation.fading import FadeIn
from manim.animation.composition import AnimationGroup
from manim.mobject.geometry import DashedLine
from manim.constants import *
from manim.mobject.geometry import Dot
from manim.utils.color import YELLOW
from manim.opengl import OpenGLMathTex
from manim._config import logger


def check_if_instance_change_if_not(obj, instance_of):
    if not isinstance(obj, instance_of):
        obj = instance_of(obj)


class ChemObject(MathTex):
    """
    `chanimlib.chem_objects.ChemObject`

    Chemical Object
    ===============

    Uses `chemfig` to create two-dimensional molecular/atomic structures.
    Only accepts `chemfig` commands.\n
    Example:\n
    >>> water = ChemObject("H_2O")    ## Linear Structure

    or

    >>> water = ChemObject("H[5]-O-H[-1]")      ## After repulsion

    You can also set various prperties of the molecule using keyword arguments. See
    `__init__` and page 7 of the manual (http://ctan.imsc.res.in/macros/generic/chemfig/chemfig-en.pdf)
    """

    # CONFIG = {"stroke_width": 2, "tex_template": ChemTemplate()}

    def __init__(
        self,
        chem_code: str,
        atom_sep: str = "2em",  ## all of these are the defaults in chemfig
        chemfig_style: str = "",
        atom_style: str = "",
        angle_increment: int = 45,
        bond_offset: str = "2pt",
        double_bond_sep: str = "2pt",
        node_style: str = "",
        bond_style: str = "",
        stroke_width: str = 2,
        tex_template=ChemTemplate,
        **kwargs,
    ):
        # digest_config(self, kwargs)
        self.template: ChemTemplate = tex_template()
        self.template.set_chemfig(
            atom_sep=atom_sep,
            chemfig_style=chemfig_style,
            atom_style=atom_style,
            angle_increment=angle_increment,
            bond_offset=bond_offset,
            double_bond_sep=double_bond_sep,
            node_style=node_style,
            bond_style=bond_style,
        )
        super().__init__(
            "\\chemfig{%s}" % chem_code,
            stroke_width=stroke_width,
            tex_template=self.template,
            **kwargs,
        )

    def set_ion_position(
        self, string_number=0, e_index=0, final_atom_index=1, direction=LEFT
    ):
        """
        This is to facilitate shifting of ionic and lewis electrons,
        by `.move_to`'ing them to the direction you provide.
        """

        self[string_number][e_index].move_to(
            self[final_atom_index].get_edge_center(direction) + direction * 0.2
        )


class OpenGLChemObject(OpenGLMathTex):
    """
    `chanimlib.chem_objects.ChemObject`

    Chemical Object
    ===============

    Uses `chemfig` to create two-dimensional molecular/atomic structures.
    Only accepts `chemfig` commands.\n
    Example:\n
    >>> water = ChemObject("H_2O")    ## Linear Structure

    or

    >>> water = ChemObject("H[5]-O-H[-1]")      ## After repulsion

    You can also set various prperties of the molecule using keyword arguments. See
    `__init__` and page 7 of the manual (http://ctan.imsc.res.in/macros/generic/chemfig/chemfig-en.pdf)
    """

    # CONFIG = {"stroke_width": 2, "tex_template": ChemTemplate()}

    def __init__(
        self,
        chem_code: str,
        atom_sep: str = "2em",  ## all of these are the defaults in chemfig
        chemfig_style: str = "",
        atom_style: str = "",
        angle_increment: int = 45,
        bond_offset: str = "2pt",
        double_bond_sep: str = "2pt",
        node_style: str = "",
        bond_style: str = "",
        stroke_width: str = 5,
        tex_template=ChemTemplate,
        **kwargs,
    ):
        # digest_config(self, kwargs)
        self.template: ChemTemplate = tex_template()
        self.template.set_chemfig(
            atom_sep=atom_sep,
            chemfig_style=chemfig_style,
            atom_style=atom_style,
            angle_increment=angle_increment,
            bond_offset=bond_offset,
            double_bond_sep=double_bond_sep,
            node_style=node_style,
            bond_style=bond_style,
        )
        super().__init__(
            "\\chemfig{%s}" % chem_code,
            stroke_width=stroke_width,
            tex_template=self.template,
            **kwargs,
        )

    def set_ion_position(
        self, string_number=0, e_index=0, final_atom_index=1, direction=LEFT
    ):
        """
        This is to facilitate shifting of ionic and lewis electrons,
        by `.move_to`'ing them to the direction you provide.
        """

        self[string_number][e_index].move_to(
            self[final_atom_index].get_edge_center(direction) + direction * 0.2
        )


class ComplexChemIon(MathTex):
    """
    Mono-ionic Complexes
    """

    # CONFIG = {"stroke_width": 2, "charge": ""}

    def __init__(
        self,
        chem_code,
        stroke_width=2,
        charge="",
        tex_template=ChemTemplate(),
        **kwargs,
    ):
        # digest_config(self, kwargs)
        self.comp = "\chemleft[\chemfig{" + chem_code + "}\chemright]^{%s}" % charge
        MathTex.__init__(
            self,
            self.comp,
            stroke_width=stroke_width,
            tex_template=tex_template,
            **kwargs,
        )


class ComplexChemCompound(MathTex):
    """
    Di-ionic Complexes
    """

    # CONFIG = {"stroke_width": 2}

    def __init__(
        self,
        cation: Union[str, ChemObject, ComplexChemIon],
        anion: Union[str, ChemObject, ComplexChemIon],
        stroke_width=2,
        **kwargs,
    ):

        if isinstance(cation, ComplexChemIon):
            cation = cation.comp
        elif isinstance(cation, ChemObject):
            cation = cation.tex_strings[0]

        if isinstance(anion, ComplexChemIon):
            anion = anion.comp
        elif isinstance(anion, ChemObject):
            anion = anion.tex_strings[0]

        MathTex.__init__(self, cation, anion, stroke_width=stroke_width, **kwargs)


## IT'S BWOKEN!! *sob*
## UPDATE 21/04/20: This... works again for some reason?
## UPDATE 07/05/20: Use Something like FadeIn for this instead of Write; because Write is fokken broken.
class Reaction(Tex):
    """
    `chanimlib.chem_objects.Reaction`
    Reaction
    ========
    Writes out a traditional chem formula

    It is recommended to use a list and not a tuple, especially if you are having only one
    reactant or product. It causes errors as python treats single items inside
    tuples as an argument of its own. So ("O=H") becomes "O","=","H".
    If you really want to use a tuple for some reason,
    add a comma at the end of the 0th item,e.g. ("O=H",)

    Arrow types:
    ============

    "forward": "->",

    "backward": "<-",

    "eq": "<=>",

    "eq_fw": "<->>",

    "eq_bw": "<<->",

    "double": "<->",

    "space": "0",

    "split": "-U"
    """

    # CONFIG = {
    #     "stroke_width": 2,
    #     # "use_hbox":false,
    #     "excluded_strings": ["\\schemestart", "\\schemestop", "\\\\"],
    #     "alignment": "",
    #     "tex_template": ChemReactionTemplate(),
    # }

    def __init__(
        self,
        reactants: List[str] = None,
        products: List[str] = None,
        arrow_type: Arrows = Arrows.forward,
        arrow_length=1,
        arrow_angle=0,
        arrow_style="",
        arrow_text_up="",
        arrow_text_down="",
        arrow_align_params="",
        debug="false",
        # use_hbox=False, idk why this is here, probably not needed now
        ## styling params from ChemObject, just keeping them here in case anyone wants to get funky with their molecule designs.
        atom_sep: str = "2em",  ## all of these are the defaults in chemfig
        chemfig_style="",
        atom_style="",
        angle_increment=45,
        bond_offset="2pt",
        double_bond_sep="2pt",
        node_style="",
        bond_style="",
        stroke_width=2,
        tex_template=ChemReactionTemplate,
        **kwargs,
    ):

        # digest_config(self, kwargs)

        # old code
        # if use_hbox:
        #     self.template_tex_file_body = TEMPLATE_CHEM_REACTION_FILE_BODY_WITH_HBOX

        self.template: ChemReactionTemplate = tex_template()
        self.template.set_chemfig(
            atom_sep=atom_sep,
            chemfig_style=chemfig_style,
            atom_style=atom_style,
            angle_increment=angle_increment,
            bond_offset=bond_offset,
            double_bond_sep=double_bond_sep,
            node_style=node_style,
            bond_style=bond_style,
            arrow_length=arrow_length,
            arrow_angle=arrow_angle,
            arrow_style=arrow_style,
            debug=debug,
        )

        if reactants is None:
            logger.warning("No reactants provided, defaulting to empty list.")
            self.reactants = []

        if products is None:
            logger.warning("No products provided, defaulting to empty list.")
            self.products = []

        if all([hasattr(reactants, "__iter__"), hasattr(reactants, "__iter__")]):
            self.reactants = reactants
            self.products = products
        else:
            raise TypeError("Reaction reactants and products must be None or iterable.")

        ## Arrow stuff
        if type(arrow_type) == Arrows:
            self.arrow_type = arrow_type.value
        elif arrow_type in Arrows.__members__.keys():  # handles old strings
            self.arrow_type = Arrows[arrow_type].value
        else:
            logger.warning(
                "Arrow type not recognised. Defaulting to Arrows.forward (->)"
            )
            self.arrow_type = Arrows.forward.value

        self.arrow_text_up = arrow_text_up
        self.arrow_text_down = arrow_text_down
        self.arrow_align_params = arrow_align_params

        self.equation = self.get_equation()
        # print(repr(self.equation))

        super().__init__(
            *self.equation,
            stroke_width=stroke_width,
            tex_template=self.template,
            **kwargs,
        )

        ##Convenience aliases.
        self.arrow = self[2 * len(self.reactants) - 1]
        self.reactants = self[: 2 * len(self.reactants) - 1 : 2]
        self.products = self[2 * len(self.reactants) :: 2]

    def get_equation(self):
        if self.arrow_align_params != "":
            arrow = "\\arrow(%s){%s[%s][%s]}" % (
                self.arrow_align_params,
                self.arrow_type,
                self.arrow_text_up,
                self.arrow_text_down,
            )
        else:
            arrow = "\\arrow{%s[%s][%s]}" % (
                self.arrow_type,
                self.arrow_text_up,
                self.arrow_text_down,
            )

        # ## To prevent manim from writing the arrow to a separate file.
        # self.excluded_strings.append(arrow)
        # print(self.excluded_strings) ##for debugging
        r = [
            "\\chemfig{" + R + "}"
            # if R != self.reactants[-1] and len(self.reactants) != 1
            # else "\\chemfig{" + R + "}"
            if "chemfig" not in R else R
            for R in self.reactants
        ]

        self.insert_in_odd_places("\+", r)

        p = [
            "\\chemfig{" + P + "}"
            # if P != self.products[-1] and len(self.products) != 1
            # else "\\chemfig{" + P + "}"
            if "chemfig" not in P else P
            for P in self.products
        ]

        self.insert_in_odd_places("\+", p)

        eq = [*r, arrow, *p]

        return eq

    def insert_in_odd_places(self, obj, iterable: list):
        """
        No pun intended.
        """

        n = len(iterable)
        # print(n)

        if n != 1:
            indexes_to_insert_at = np.array(range(1, n))
            print(indexes_to_insert_at)

            for i in indexes_to_insert_at:
                iterable.insert(-i, obj)
                print(iterable)
                indexes_to_insert_at += 1

    # I think this was intended to give a dict or something of all reactants and products. IDK
    def get_breakdown_dict(self):
        pass

    # can't really think of a better name for this atm, and show will overwrite Mobject.show
    def creation_anim(
        self,
        text_anim: Animation = Write,
        arrow_anim: Animation = FadeIn,
        reactant_product_simultaneity=False,
        text_anim_kwargs: dict = None,
        arrow_anim_kwargs: dict = None,
        group_anim_kwargs: dict = None,
    ) -> AnimationGroup:
        """Workaround and shortcut method to overcome the bugs in `Write`.

        Args:
            text_anim (Animation, optional): The animation on the reactants and products. Defaults to Write.
            arrow_anim (Animation, optional): The animation on the arrow. Defaults to FadeIn.
            reactant_product_simultaneity (bool, optional): Whether to animate the reactants and products together or not.
        Returns:
            AnimationGroup: The group of animations on the text and arrow.
        """
        text = VGroup(
            self[0 : 2 * len(self.reactants) - 1], self[2 * len(self.reactants) :]
        )
        arrow = self[2 * len(self.reactants) - 1]

        if text_anim_kwargs is None:
            text_anim_kwargs = dict()

        if arrow_anim_kwargs is None:
            arrow_anim_kwargs = dict()

        if group_anim_kwargs is None:
            group_anim_kwargs = dict()

        anim_group = (
            AnimationGroup(
                text_anim(text[0], **text_anim_kwargs),
                text_anim(text[1], **text_anim_kwargs),
                arrow_anim(arrow, **arrow_anim_kwargs),
                **group_anim_kwargs,
            )
            if reactant_product_simultaneity
            else AnimationGroup(
                text_anim(text, **text_anim_kwargs),
                arrow_anim(arrow, **arrow_anim_kwargs),
                **group_anim_kwargs,
            )
        )

        return anim_group


class ChemArrow(MathTex):
    """
    Chemical Reaction Arrow
    =======================
    Basically a Reaction without any reactants/products.

    Arguments:
    ----------

    mode: Arrows - The type of arrow to be used. See :class:`Arrows` for more info on available options.

    length - length of the arrow. Sets `arrow coeff` chemfig property.

    angle - angle made with respect to the x-axis. Sets `arrow angle` chemfig property.

    style - Single string of comma separated TiKZ styling parameters. Sets `arrow style` chemfig property.

    text_up - text to be displayed above the main arrow line.

    text_down - text to be displayed below the main arrow line.


    """

    def __init__(
        self,
        mode=Arrows.forward,
        length=1,
        angle=0,
        style="",
        text_up="",
        text_down="",
        stroke_width=2,
        tex_template=ChemReactionTemplate,
        **kwargs,
    ):
        self.tex_template: ChemReactionTemplate = tex_template()

        self.tex_template.set_chemfig(
            arrow_angle=angle, arrow_length=length, arrow_style=style
        )
        arrow = "\\arrow{%s[%s][%s]}" % (mode.value, text_up, text_down)
        MathTex.__init__(
            self,
            arrow,
            stroke_width=stroke_width,
            tex_template=self.tex_template,
            **kwargs,
        )


class ChemWithName(VMobject):
    """
    `chanimlib.chem_objects.ChemWithName`
    Chemical With Name
    ==================
    A better version of `ChemName` that allows you to animate
    either part.

    To animate:
    -----------
    ```py
    ChemWithName.creation_anim(chem_anim,name_anim)
    ```


    inspired by `BraceLabel`
    """


    def __init__(
        self, chem, name, name_direction=DOWN, label_constructor=Tex, buff=1, **kwargs
    ):

        super().__init__(self, **kwargs)

        if isinstance(
            chem, (ChemObject, ComplexChemIon, ComplexChemCompound, ChemAbove)
        ):
            self.chem = chem
        else:
            self.chem = ChemObject(chem, **kwargs)

        if isinstance(name, label_constructor):
            self.name = name
        else:
            self.name: Tex = label_constructor(name, **kwargs)

        self.name.next_to(self.chem, name_direction, buff=buff)

        self.submobjects = [self.chem, self.name]

    def creation_anim(
        self,
        chem_anim: Animation = Write,
        name_anim: Animation = FadeIn,
        chem_anim_kwargs=None,
        name_anim_kwargs=None,
        group_anim_kwargs=None,
    ):
        if chem_anim_kwargs is None:
            chem_anim_kwargs = dict()

        if name_anim_kwargs is None:
            name_anim_kwargs = dict()

        if group_anim_kwargs is None:
            group_anim_kwargs = dict()

        return AnimationGroup(
            chem_anim(self.chem, **chem_anim_kwargs),
            name_anim(self.name, **name_anim_kwargs),
            **group_anim_kwargs,
        )


class ChemAbove(ChemObject):
    def __init__(self, chem, above_chem, hspace="3mm", vspace="1mm", **kwargs):
        ChemObject.__init__(
            self,
            "\\chemabove{%s}{\\hspace{%s} \\vspace{%s} %s}"
            % (chem, hspace, vspace, above_chem),
            **kwargs,
        )


# Doesn't work, somebody fix this or I'll say it's deprecated.


class ReactionVGroup(VGroup):
    CONFIG = {
        "_type": "forward",
        "length": 1,
        "angle": 0,
        "style": "{}",
        "text_up": "",
        "text_down": "",
    }

    def __init__(
        self,
        reactants: List[ChemObject] = [],
        products: List[ChemObject] = [],
        **arrow_kwargs,
    ):
        raise DeprecationWarning(
            "ReactionVGroup doesn't work, and will probably be removed in the future"
        )
        digest_config(self, arrow_kwargs)
        arrow = ChemArrow(
            self._type,
            self.length,
            self.angle,
            self.style,
            self.text_up,
            self.text_down,
        )

        r = self.get_side_of_equation(reactants)
        p = self.get_side_of_equation(products)

        self.set_same_height_for_all_mobjects(r)
        self.set_same_height_for_all_mobjects(p)
        self.set_same_height_for_all_mobjects([VGroup(*r), arrow, VGroup(*p)])

        VGroup.__init__(*[*r, arrow, *p])

    def get_side_of_equation(self, iterableable):
        # A = [
        #     *(Participant, MathTex("+")) if Participant != iterableable[-1]
        #     else Participant
        #     for Participant in iterableable
        # ]

        A = []
        for Part in iterableable:
            if Part != iterableable[-1]:
                A.extend([ChemObject(Part), MathTex("+")])
            else:
                A.append(ChemObject(Part))
        return A

    def set_same_height_for_all_mobjects(self, iterableable):
        for index in range(len(iterableable)):
            if index != 0:
                iterableable[index].set_y(iterableable[index - 1].get_y())


class BondBreak(DashedLine):
    """
    `chanimlib.chem_objects.BondBreak`

    Bond Break
    ----------

    A line that splits a given ``bond``.
    """

    # CONFIG = {
    #     "length": 0.7,
    # }

    def __init__(self, bond: TexSymbol, length=0.7, **kwargs):
        # digest_config(self, kwargs)
        start = bond.get_center() + (length / 2) * DOWN
        end = start + length * UP

        DashedLine.__init__(self, start=start, end=end)


class ElectronPair(VGroup):
    """Electron Pair: Two electrons (Dots) in a VGroup

    Arguments:
        None -- Just use this as it is and change the kwargs if you want
    """

    # CONFIG = dict(color=YELLOW, pair_buff=0.15)

    def __init__(self, color=YELLOW, pair_buff=0.15, **kwargs):
        super().__init__(
            Dot(radius=DEFAULT_SMALL_DOT_RADIUS),
            Dot(RADIUS=DEFAULT_SMALL_DOT_RADIUS),
            **kwargs,
        )
        self.arrange(RIGHT, buff=pair_buff).set_color(color)


## Aliases
Formula = ChemObject
CArrow = ChemArrow
React = Reaction



from .chem_objects import ChemObject

"""
Lazy Classes to create useful organic compounds.
"""

class PrimaryAlcohol(ChemObject):
    """
    Returns a straight chain alcohol.
    chain_length >= 1

    TODO: Isn't working currently for some reason.
          Will fix later.
    """
    def __init__(self, chain_length=1):

        compound = methyl + (chain_length-1)*methylene + "-OH"
        print(compound)

        ChemObject.__init__(self,
                            compound)


class SecondaryAlcohol(ChemObject):
    """
    Use pos >= 2  and chain_length >= 3 for desired results.
    """

    def __init__(self, pos=2, chain_length=3):
        ChemObject.__init__(self,
                            methyl + (pos-1)*methylene +
                            methine + "(-OH)" +
                            (chain_length-pos-1)*methylene + methyl)


class PrimaryAcid(ChemObject):
    """
    Returns a straight chain carboxylic acid.
    chain_length >= 1
    """
    def __init__(self, chain_length=1):
        if chain_length == 1:
            ChemObject.__init__(self,
                                "H[3]-C(-OH[5])=O")
        else:
            compound = methyl + (chain_length-2)*methylene + "-COOH"
            print(compound)
            ChemObject.__init__(self,
                                compound)

"""
These are a few compounds that are very frequently used and
so can be called directly, without having to write the long chemfig code
again and again. Saves time, really.
"""

methyl = "[4]H-C(-[2]H)(-[6]H)"

methylene = "-C(-[2]H)(-[6]H)"

methine = "-C(-[2]H)"

Phenol = "*6(-=-=(-OH)-=)"

Water = "[5]H-\lewis{1:2:,O}-H[-1]"

# Ethanol = PrimaryAlcohol(2)

# Acetic_Acid = Ethanoic_Acid = PrimaryAcid(2)

Benzene = "*6(-=-=-=)"


Benzene_Diazonium_Chloride = BDC = "*6(-=-=(-\\chemabove{N_2}{\quad\scriptstyle+}\\chemabove{Cl}{\quad\scriptstyle-})-=-)"

Aniline ="*6(-=-=(-NH2)-=-)"

Carbon_Dioxide = CO2 = "[4]O=C=O"


## Bonds that can be a pain in the arse

def cbond(
          _dir="->",
          angle=0,
          coeff=1.2,
          n1="",
          n2=""
        ) -> str:
    """func cbond

    Return the code for a coordinate/dative bond

    Args:
        _dir : Direction of the arrow
        angle: Angle from baseline
        coeff: Bond length multiplier
        n1/n2: Atom identifiers

    Returns:
        A string with the list of all this stuff.
    """
    # bond = []
    # for param in params:
    #     if param
    # arrow = "->"
    return f"[{angle},{coeff},{n1},{n2},{_dir}]"

CBL=C_BOND_LEFT = COORDINATE_BOND_LEFT = cbond("<-")
CBR=C_BOND_RIGHT = COORDINATE_BOND_RIGHT = cbond("->")



from manim.animation.transform import Transform, ReplacementTransform

class BondToElectronPair(Transform):
    def __init__(self,mob,bond_pos):
        bond_with_atoms = mob[0][bond_pos-2:band_pos+1]
        
        atom_2 = mob[0][bond_pos-1]
        atom_1 = mob[0][bond_pos-2]

        # e_pair = self.get_pair_as_per_bond_orientation(bond)
        e_pair = "\\lewis"
        Transform.__init__(bond,e_pair)

    def get_pair_as_per_bond_orientation(self,bond):
        pass


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chanim']

package_data = \
{'': ['*']}

install_requires = \
['manim']

entry_points = \
{'manim.plugins': ['chanim = chanim']}

setup_kwargs = {
    'name': 'chanim',
    'version': '1.0.2.1',
    'description': 'Manim extension for making chemistry videos',
    'long_description': '# Chanim\nThis is an extension to [Manim](https://www.github.com/ManimCommunity/manim) library (initially created by [3Blue1Brown](https://github.com/3b1b/manim)),\nfor making videos regarding chemistry.\n\n> A Hindi version of this README is available [here](https://github.com/raghavg123/chanim/blob/master/README-%E0%A4%B9%E0%A4%BF%E0%A4%A8%E0%A5%8D%E0%A4%A6%E0%A5%80.md).\n\n## Installation (pip)\n`pip install chanim`\n\n## Installation (Source)\n1. Install the external dependencies for manim as described [here](https://docs.manim.community/en/latest/installation.html) according to your OS.\n2. Clone the contents of this repository.\n3. Open a terminal in the cloned directory and run `pip install -e .`, or if you prefer to use [poetry](https://python-poetry.org) instead, `poetry install`. This\'ll install `manim` for you as well if you don\'t already have it installed. (you\'ll still need to setup the external dependencies though)\n\nThat\'s about it. You can now do `from chanim import <*|object_name>` like any regular Python package. \n\n## Usage\n\nHere\'s a little example of it working.\n\n```py\nfrom chanim import *\n\nclass ChanimScene(Scene):\n    def construct(self):\n        ## ChemWithName creates a chemical diagram with a name label\n        chem = ChemWithName("*6((=O)-N(-CH_3)-*5(-N=-N(-CH_3)-=)--(=O)-N(-H_3C)-)", "Caffeine")\n\n        self.play(chem.creation_anim())\n        self.wait()\n```\n\nType this into a python (`.py`) file. I\'ll assume you named it `chem.py`\n\nIn your command prompt/terminal write this (assuming you\'re in your project directory):\n\n```sh\nmanim -p -qm chem.py ChanimScene\n```\nThis\'ll render your Scene and `p`review it in your default player (in `m`edium `q`uality).\n\n\nhttps://user-images.githubusercontent.com/65204531/124297601-dcafcf80-db78-11eb-936b-cdc913c91f25.mp4\n\n\nCongrats! You\'ve written and played your first animation with chanim (or "chanimation" should I say)\n\nExplore the code and docs (coming soon!) for more on how to use chanim.\n\n## Abilities\nCurrently chanim only supports drawing compounds and reactions along with a few chemfig commands (such as coordinate bonds and complexes etc.) but more is to come! If you have a suggestion, file an issue with a proper tag.\n\n## A Quick Note\nThere may be some faulty code and a lot of this may not be well made/documented. Feel free to file an issue if something doesn\'t work properly.\n',
    'author': 'kilacoda',
    'author_email': 'kilacoda@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/raghavg123/chanim',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""A few TexTemplate subclasses used internally by chanim but can be used by a user if they want to as well.
Note that setting the TeX template for a scene will affect all subsequent TexMobjects until another change in template.
"""

from manim.utils.tex import TexTemplate


class ChemTemplate(TexTemplate):
    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        self.add_to_preamble("\\usepackage{chemfig}")


    def set_chemfig(
        self,
        atom_sep: str = "2em", ## all of these are the defaults in chemfig
        chemfig_style:str="",
        atom_style:str="",
        angle_increment:int=45,
        bond_offset:str="2pt",
        double_bond_sep:str="2pt",
        node_style:str="",
        bond_style:str="",
    ):
        set_chemfig = "\\setchemfig{atom sep=%s,chemfig style=%s, atom style=%s,angle increment=%d,bond offset=%s,double bond sep=%s, node style=%s, bond style=%s}" % (atom_sep,chemfig_style,atom_style,angle_increment,bond_offset,double_bond_sep,node_style,bond_style)
        self.add_to_preamble(set_chemfig)


class ChemReactionTemplate(TexTemplate):
    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        self.add_to_preamble("\\usepackage{chemfig}")

    def set_chemfig(
        self,

        ##Individual molecule params
        atom_sep: str = "2em", ## all of these are the defaults in chemfig
        chemfig_style:str="",
        atom_style:str="",
        angle_increment:int=45,
        bond_offset:str="2pt",
        double_bond_sep:str="2pt",
        node_style:str="",
        bond_style:str="",

        ## Reaction scheme params
        arrow_length:int=1,
        arrow_angle:int=0,
        arrow_style:int="",
        debug:str="false",
    ):
        set_chemfig = "\\setchemfig{atom sep=%s,chemfig style=%s, atom style=%s,angle increment=%d,bond offset=%s,double bond sep=%s, node style=%s, bond style=%s,scheme debug=%s, atom sep=2em, arrow angle={%d}, arrow coeff={%s}, arrow style={%s}}" % (atom_sep,chemfig_style,atom_style,angle_increment,bond_offset,double_bond_sep,node_style,bond_style,debug, arrow_angle, arrow_length, arrow_style)

        self.add_to_preamble(set_chemfig)

    def get_texcode_for_expression(self, expression):
        """Inserts expression verbatim into TeX template.

        Parameters
        ----------
        expression : :class:`str`
            The string containing the expression to be typeset, e.g. ``$\\sqrt{2}$``

        Returns
        -------
        :class:`str`
            LaTeX code based on current template, containing the given ``expression`` and ready for typesetting
        """
        return self.body.replace(self.placeholder_text, "\n\\schemestart\n"+expression+"\n\\schemestop\n\n")

    def get_texcode_for_expression_in_env(self,expression,environment):
        """Inserts an expression wrapped in a given environment into the TeX template.

        Parameters
        ----------
        environment : :class:`str`
            The environment in which we should wrap the expression.
        expression : :class:`str`
            The string containing the expression to be typeset, e.g. ``"$\\sqrt{2}$"``

        Returns
        -------
        :class:`str`
            LaTeX code based on template, containing the given expression and ready for typesetting
        """
        print(environment)
        begin = r"\begin{" + environment + "}" + "\n\\schemestart"
        end = "\\schemestop\n" + r"\end{" + environment + "}"
        print(begin,end)
        return self.body.replace(
            self.placeholder_text, "{0}\n{1}\n{2}".format(begin,expression, end)
        )



from enum import Enum

__all__ = ["Arrows"]


class Arrows(Enum):
    """
    An enum of chemfig-recognised arrow types.
    See p. 49 of http://ctan.imsc.res.in/macros/generic/chemfig/chemfig-en.pdf for
    visual examples.
    """

    forward = "->"
    backward = "<-"
    eq = "<=>"
    eq_fw = "<->>"
    eq_bw = "<<->"
    double = "<->"
    space = "0"
    split = "-U"
